# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import logging
import warnings
from abc import abstractmethod
from copy import deepcopy
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

import earthkit.data as ekd
import numpy as np
from earthkit.data.core.temporary import temp_file
from earthkit.data.readers.grib.output import new_grib_output

from anemoi.datasets.create.sources.mars import mars
from anemoi.datasets.create.utils import to_datetime_list

from .legacy import legacy_source

LOG = logging.getLogger(__name__)


xprint = print


def _member(field: Any) -> int:
    """Retrieves the member number from the field metadata.

    Parameters
    ----------
    field : Any
        The field from which to retrieve the member number.

    Returns
    -------
    int
        The member number.
    """
    # Bug in eccodes has number=0 randomly
    number = field.metadata("number", default=0)
    if number is None:
        number = 0
    return number


class Period:
    value = None

    def __init__(self, start_datetime, end_datetime, base_datetime):
        assert isinstance(start_datetime, datetime.datetime)
        assert isinstance(end_datetime, datetime.datetime)
        assert isinstance(base_datetime, datetime.datetime)

        self.start_datetime = start_datetime
        self.end_datetime = end_datetime

        self.base_datetime = base_datetime

    @property
    def time_request(self):
        date = int(self.base_datetime.strftime("%Y%m%d"))
        time = int(self.base_datetime.strftime("%H%M"))

        end_step = self.end_datetime - self.base_datetime
        assert end_step.total_seconds() % 3600 == 0, end_step  # only full hours supported
        end_step = int(end_step.total_seconds() // 3600)

        return (("date", date), ("time", time), ("step", end_step))

    def field_to_key(self, field):
        return (
            ("date", field.metadata("date")),
            ("time", field.metadata("time")),
            ("step", field.metadata("step")),
        )

    def check(self, field):
        stepType = field.metadata("stepType")
        startStep = field.metadata("startStep")
        endStep = field.metadata("endStep")
        date = field.metadata("date")
        time = field.metadata("time")

        assert stepType == "accum", stepType

        base_datetime = datetime.datetime.strptime(str(date) + str(time).zfill(4), "%Y%m%d%H%M")

        start = base_datetime + datetime.timedelta(hours=startStep)
        assert start == self.start_datetime, (start, self.start_datetime)

        end = base_datetime + datetime.timedelta(hours=endStep)
        assert end == self.end_datetime, (end, self.end_datetime)

    def is_matching_field(self, field):
        return self.field_to_key(field) == self.time_request

    def __repr__(self):
        return f"Period({self.start_datetime} to {self.end_datetime} -> {self.time_request})"

    def length(self):
        return self.end_datetime - self.start_datetime

    def apply(self, accumulated, values):

        if accumulated is None:
            accumulated = np.zeros_like(values)

        assert accumulated.shape == values.shape, (accumulated.shape, values.shape)

        if not np.all(values >= 0):
            warnings.warn(f"Negative values for {values}: {np.amin(values)} {np.amax(values)}")

        return accumulated + self.sign * values


class TodoList:
    def __init__(self, keys):
        self._todo = set(keys)
        self._len = len(keys)
        self._done = set()
        assert self._len == len(self._todo), (self._len, len(self._todo))

    def is_todo(self, key):
        return key in self._todo

    def is_done(self, key):
        return key in self._done

    def set_done(self, key):
        self._done.add(key)
        self._todo.remove(key)

    def all_done(self):
        if not self._todo:
            assert len(self._done) == self._len, (len(self._done), self._len)
            return True
        return False


class Periods:
    _todo = None

    def __init__(self, valid_date, accumulation_period, **kwargs):
        # one Periods object for each accumulated field in the output

        assert isinstance(valid_date, datetime.datetime), (valid_date, type(valid_date))
        assert isinstance(accumulation_period, datetime.timedelta), (accumulation_period, type(accumulation_period))
        self.valid_date = valid_date
        self.accumulation_period = accumulation_period
        self.kwargs = kwargs

        self._periods = self.build_periods()
        self.check_merged_interval()

    def check_merged_interval(self):
        global_start = self.valid_date - self.accumulation_period
        global_end = self.valid_date
        resolution = datetime.timedelta(hours=1)

        timeline = np.arange(
            np.datetime64(global_start, "s"), np.datetime64(global_end, "s"), np.timedelta64(resolution)
        )

        flags = np.zeros_like(timeline, dtype=int)
        for p in self._periods:
            segment = np.where((timeline >= p.start_datetime) & (timeline < p.end_datetime))
            xprint(segment)
            flags[segment] += p.sign
        assert np.all(flags == 1), flags

    def find_matching_period(self, field):
        # Find a period that matches the field, or return None
        found = [p for p in self._periods if p.is_matching_field(field)]
        if len(found) == 1:
            return found[0]
        if len(found) > 1:
            raise ValueError(f"Found more than one period for {field}")
        return None

    @property
    def todo(self):
        if self._todo is None:
            self._todo = TodoList([p.time_request for p in self._periods])
        return self._todo

    def is_todo(self, period):
        return self.todo.is_todo(period.time_request)

    def is_done(self, period):
        return self.todo.is_done(period.time_request)

    def set_done(self, period):
        self.todo.set_done(period.time_request)

    def all_done(self):
        return self.todo.all_done()

    def __iter__(self):
        return iter(self._periods)

    @abstractmethod
    def build_periods(self):
        pass


class EraPeriods(Periods):
    def search_periods(self, start, end, debug=False):
        # find candidate periods that can be used to accumulate the data
        # to get the accumulation between the two dates 'start' and 'end'
        found = []
        if not end - start == datetime.timedelta(hours=1):
            raise NotImplementedError("Only 1 hour period is supported")

        for base_time, steps in self.available_steps(start, end).items():
            for step1, step2 in steps:
                if debug:
                    xprint(f"❌ tring: {base_time=} {step1=} {step2=}")

                if ((base_time + step1) % 24) != start.hour:
                    continue

                if ((base_time + step2) % 24) != end.hour:
                    continue

                base_datetime = start - datetime.timedelta(hours=step1)

                period = Period(start, end, base_datetime)
                found.append(period)

                assert base_datetime.hour == base_time, (base_datetime, base_time)

                assert period.start_datetime - period.base_datetime == datetime.timedelta(hours=step1), (
                    period.start_datetime,
                    period.base_datetime,
                    step1,
                )
                assert period.end_datetime - period.base_datetime == datetime.timedelta(hours=step2), (
                    period.end_datetime,
                    period.base_datetime,
                    step2,
                )

        return found

    def build_periods(self):
        # build the list of periods to accumulate the data

        hours = self.accumulation_period.total_seconds() / 3600
        assert int(hours) == hours, f"Only full hours accumulation is supported {hours}"
        hours = int(hours)

        lst = []
        for wanted in [[i, i + 1] for i in range(0, hours, 1)]:

            start = self.valid_date - datetime.timedelta(hours=wanted[1])
            end = self.valid_date - datetime.timedelta(hours=wanted[0])

            found = self.search_periods(start, end)
            if not found:
                xprint(f"❌❌❌ Cannot find accumulation for {start} {end}")
                self.search_periods(start, end, debug=True)
                raise ValueError(f"Cannot find accumulation for {start} {end}")

            found = sorted(found, key=lambda x: x.base_datetime, reverse=True)
            chosen = found[0]

            if len(found) > 1:
                xprint(f"  Found more than one period for {start} {end}")
                for f in found:
                    xprint(f"    {f}")
                xprint(f"    Chosing {chosen}")

            chosen.sign = 1

            lst.append(chosen)
        return lst


class EaOperPeriods(EraPeriods):
    def available_steps(self, start, end):
        return {
            6: [[i, i + 1] for i in range(0, 18, 1)],
            18: [[i, i + 1] for i in range(0, 18, 1)],
        }


class L5OperPeriods(EraPeriods):
    def available_steps(self, start, end):
        print("❌❌❌ untested")
        x = 24  # need to check if 24 is the right value
        return {
            0: [[i, i + 1] for i in range(0, x, 1)],
        }


class EaEndaPeriods(EraPeriods):
    def available_steps(self, start, end):
        print("❌❌❌ untested")
        return {
            6: [[i, i + 3] for i in range(0, 18, 1)],
            18: [[i, i + 3] for i in range(0, 18, 1)],
        }


class RrOperPeriods(Periods):
    def available_steps(self, start, end):
        raise NotImplementedError("need to implement diff")
        x = 24  # todo: check if 24 is the right value
        return {
            0: [[0, i] for i in range(0, x, 1)],
            3: [[0, i] for i in range(0, x, 1)],
            6: [[0, i] for i in range(0, x, 1)],
            9: [[0, i] for i in range(0, x, 1)],
            12: [[0, i] for i in range(0, x, 1)],
            15: [[0, i] for i in range(0, x, 1)],
            18: [[0, i] for i in range(0, x, 1)],
            21: [[0, i] for i in range(0, x, 1)],
        }


class OdEldaPeriods(EraPeriods):
    def available_steps(self, start, end):
        print("❌❌❌ untested")
        x = 24  # need to check if 24 is the right value
        return {
            6: [[i, i + 1] for i in range(0, x, 1)],
            18: [[i, i + 1] for i in range(0, x, 1)],
        }


class DiffPeriods(Periods):
    pass


class OdOperPeriods(DiffPeriods):
    def available_steps(self, start, end):
        raise NotImplementedError("need to implement diff and _scda patch")


class OdEnfoPeriods(DiffPeriods):
    def available_steps(self, start, end):
        raise NotImplementedError("need to implement diff")


def find_accumulator_class(class_: str, stream: str) -> Periods:
    return {
        ("ea", "oper"): EaOperPeriods,  # runs ok
        ("ea", "enda"): EaEndaPeriods,
        ("rr", "oper"): RrOperPeriods,
        ("l5", "oper"): L5OperPeriods,
        ("od", "oper"): OdOperPeriods,
        ("od", "enfo"): OdEnfoPeriods,
        ("od", "elda"): OdEldaPeriods,
    }[class_, stream]


class Accumulator:
    values = None

    def __init__(self, period_class, out, valid_date, user_accumulation_period, **kwargs):
        self.valid_date = valid_date

        # keep the reference to the output file to be able to write the result using an input field as template
        self.out = out

        # key contains the mars request parameters except the one related to the time
        # A mars request is a dictionary with three categories of keys:
        #   - the ones related to the time (date, time, step)
        #   - the ones related to the data (param, stream, levtype, expver, number, ...)
        #   - the ones related to the processing to be done (grid, area, ...)
        self.kwargs = kwargs
        for k in ["date", "time", "step"]:
            if k in kwargs:
                raise ValueError(f"Cannot use {k} in kwargs for accumulations")

        self.key = {k: v for k, v in kwargs.items() if k in ["param", "level", "levelist", "number"]}

        self.periods = period_class(self.valid_date, user_accumulation_period, **kwargs)

    @property
    def requests(self):
        for period in self.periods:
            # build the full data requests, merging the time requests with the key
            yield {**self.kwargs.copy(), **dict(period.time_request)}

    def is_field_needed(self, field):
        for k, v in self.key.items():
            if field.metadata(k) != v:
                LOG.debug(f"{self} does not need field {field} because of {k}={field.metadata(k)} not {v}")
                return False
        return True

    def compute(self, field, values):
        if not self.is_field_needed(field):
            return

        period = self.periods.find_matching_period(field)
        if not period:
            return
        assert self.periods.is_todo(period), (self.periods, period)
        assert not self.periods.is_done(period), f"Field {field} for period {period} already done"

        period.check(field)

        xprint(f"{self}  field ✅ ({period.sign}){field} for {period}")

        self.values = period.apply(self.values, values)
        self.periods.set_done(period)

        if self.periods.all_done():
            self.write(field)
            xprint("accumulator", self, " : data written ✅ ")

    def check(self, field: Any) -> None:
        if self._check is None:
            self._check = field.metadata(namespace="mars")

            assert self.param == field.metadata("param"), (self.param, field.metadata("param"))
            assert self.date == field.metadata("date"), (self.date, field.metadata("date"))
            assert self.time == field.metadata("time"), (self.time, field.metadata("time"))
            assert self.step == field.metadata("step"), (self.step, field.metadata("step"))
            assert self.number == _member(field), (self.number, _member(field))
            return

        mars = field.metadata(namespace="mars")
        keys1 = sorted(self._check.keys())
        keys2 = sorted(mars.keys())

        assert keys1 == keys2, (keys1, keys2)

        for k in keys1:
            if k not in ("step",):
                assert self._check[k] == mars[k], (k, self._check[k], mars[k])

    def write(self, template: Any) -> None:
        assert self.periods.all_done(), self.periods

        if np.all(self.values < 0):
            LOG.warning(
                f"Negative values when computing accumutation for {self}): min={np.amin(self.values)} max={np.amax(self.values)}"
            )

        startStep = 0
        endStep = self.periods.accumulation_period.total_seconds() // 3600
        assert int(endStep) == endStep, "only full hours accumulation is supported"
        endStep = int(endStep)
        fake_base_date = self.valid_date - self.periods.accumulation_period
        date = int(fake_base_date.strftime("%Y%m%d"))
        time = int(fake_base_date.strftime("%H%M"))

        self.out.write(
            self.values,
            template=template,
            stepType="accum",
            startStep=startStep,
            endStep=endStep,
            date=date,
            time=time,
            check_nans=True,
        )
        self.values = None

    def __repr__(self):
        key = ", ".join(f"{k}={v}" for k, v in self.key.items())
        return f"{self.__class__.__name__}({self.valid_date}, {key})"


def _compute_accumulations(
    context: Any,
    dates: List[datetime.datetime],
    request: Dict[str, Any],
    user_accumulation_period: datetime.timedelta,
    # data_accumulation_period: Optional[int] = None,
    # patch: Any = _identity,
) -> Any:

    request = deepcopy(request)

    param = request.pop("param")
    assert isinstance(param, (list, tuple))

    number = request.pop("number", [0])
    if not isinstance(number, (list, tuple)):
        number = [number]
    assert isinstance(number, (list, tuple))

    request["stream"] = request.get("stream", "oper")

    type_ = request.get("type", "an")
    if type_ == "an":
        type_ = "fc"
    request["type"] = type_

    request["levtype"] = request.get("levtype", "sfc")
    if request["levtype"] != "sfc":
        # LOG.warning("'type' should be 'sfc', found %s", request['type'])
        raise NotImplementedError("Only sfc leveltype is supported")

    period_class = find_accumulator_class(request["class"], request["stream"])

    tmp = temp_file()
    path = tmp.path
    out = new_grib_output(path)

    # build one accumulator per output field
    accumulators = []
    for valid_date in dates:
        for p in param:
            for n in number:
                accumulators.append(
                    Accumulator(
                        period_class,
                        out,
                        valid_date,
                        user_accumulation_period=user_accumulation_period,
                        param=p,
                        number=n,
                        **request,
                    )
                )

    xprint("accumulators", len(accumulators))

    # get all needed data requests (mars)
    requests = []
    for a in accumulators:
        xprint("accumulator", a)
        for r in a.requests:
            xprint(" ", r)
            requests.append(r)

    # get the data (this will pack the requests to avoid duplicates and make a minimal number of requests)
    ds = mars(context, dates, request_already_using_valid_datetime=True, *requests)

    # send each field to the each accumulator, the accumulatore will use the field to the accumulation
    # if the accumulator has requested it
    for field in ds:
        values = field.values  # optimisation
        for a in accumulators:
            a.compute(field, values)

    out.close()

    ds = ekd.from_source("file", path)

    assert len(ds) / len(param) / len(number) == len(dates), (
        len(ds),
        len(param),
        len(dates),
    )

    # keep a reference to the tmp file, or it gets deleted when the function returns
    ds._tmp = tmp

    return ds


def _to_list(x: Union[List[Any], Tuple[Any], Any]) -> List[Any]:
    """Converts the input to a list if it is not already a list or tuple.

    Parameters
    ----------
    x : Union[List[Any], Tuple[Any], Any]
        Input value.

    Returns
    -------
    List[Any]
        The input value as a list.
    """
    if isinstance(x, (list, tuple)):
        return x
    return [x]


def _scda(request: Dict[str, Any]) -> Dict[str, Any]:
    """Modifies the request stream based on the time.

    Parameters
    ----------
    request : Dict[str, Any]
        Request parameters.

    Returns
    -------
    Dict[str, Any]
        The modified request parameters.
    """
    if request["time"] in (6, 18, 600, 1800):
        request["stream"] = "scda"
    else:
        request["stream"] = "oper"
    return request


@legacy_source(__file__)
def accumulations(context, dates, **request):
    _to_list(request["param"])
    user_accumulation_period = request.pop("accumulation_period", 6)
    user_accumulation_period = datetime.timedelta(hours=user_accumulation_period)

    context.trace("🌧️", f"accumulations {request} {user_accumulation_period}")

    return _compute_accumulations(
        context,
        dates,
        request,
        user_accumulation_period=user_accumulation_period,
    )


execute = accumulations

if __name__ == "__main__":
    import yaml

    config = yaml.safe_load(
        """
      class: ea
      expver: '0001'
      grid: 20./20.
      levtype: sfc
#      number: [0, 1]
#      stream: enda
      param: [cp, tp]
#      accumulation_period: 6h
      accumulation_period: 2
    """
    )
    dates = yaml.safe_load("[2022-12-31 00:00, 2022-12-31 06:00]")
    # dates = yaml.safe_load("[2022-12-30 18:00, 2022-12-31 00:00, 2022-12-31 06:00, 2022-12-31 12:00]")
    dates = to_datetime_list(dates)

    class Context:
        use_grib_paramid = True

        def trace(self, *args):
            print(*args)

    for f in accumulations(Context, dates, **config):
        print(f, f.to_numpy().mean())



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ..grids import cropping_mask
from .dataset import Dataset
from .dataset import FullIndex
from .dataset import Shape
from .dataset import TupleIndex
from .debug import Node
from .debug import debug_indexing
from .forwards import Forwards
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import update_tuple

LOG = logging.getLogger(__name__)


class Masked(Forwards):
    """A class to represent a masked dataset."""

    def __init__(self, forward: Dataset, mask: NDArray[np.bool_]) -> None:
        """Initialize the Masked class.

        Parameters
        ----------
        forward : Dataset
            The dataset to be masked.
        mask : NDArray[np.bool_]
            The mask array.
        """
        super().__init__(forward)
        assert len(forward.shape) == 4, "Grids must be 1D for now"
        self.mask = mask
        self.axis = 3

        self.mask_name = f"{self.__class__.__name__.lower()}_mask"

    @cached_property
    def shape(self) -> Shape:
        """Get the shape of the masked dataset."""
        return self.forward.shape[:-1] + (np.count_nonzero(self.mask),)

    @cached_property
    def latitudes(self) -> NDArray[Any]:
        """Get the masked latitudes."""
        return self.forward.latitudes[self.mask]

    @cached_property
    def longitudes(self) -> NDArray[Any]:
        """Get the masked longitudes."""
        return self.forward.longitudes[self.mask]

    @debug_indexing
    def __getitem__(self, index: FullIndex) -> NDArray[Any]:
        """Get the masked data at the specified index.

        Parameters
        ----------
        index : FullIndex
            The index to retrieve data from.

        Returns
        -------
        NDArray[Any]
            The masked data at the specified index.
        """
        if isinstance(index, tuple):
            return self._get_tuple(index)

        result = self.forward[index]
        # We don't support subsetting the grid values
        assert result.shape[-1] == len(self.mask), (result.shape, len(self.mask))

        return result[..., self.mask]

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        """Get the masked data for a tuple index.

        Parameters
        ----------
        index : TupleIndex
            The tuple index to retrieve data from.

        Returns
        -------
        NDArray[Any]
            The masked data for the tuple index.
        """
        index, changes = index_to_slices(index, self.shape)
        index, previous = update_tuple(index, self.axis, slice(None))
        result = self.forward[index]
        result = result[..., self.mask]
        result = result[..., previous]
        result = apply_index_to_slices_changes(result, changes)
        return result

    def collect_supporting_arrays(self, collected: List[Tuple], *path: Any) -> None:
        """Collect supporting arrays.

        Parameters
        ----------
        collected : List[Tuple]
            The list to collect supporting arrays into.
        path : Any
            Additional path arguments.
        """
        super().collect_supporting_arrays(collected, *path)
        collected.append((path, self.mask_name, self.mask))


class Thinning(Masked):
    """A class to represent a thinned dataset."""

    def __init__(self, forward: Dataset, thinning: Optional[int], method: str) -> None:
        """Initialize the Thinning class.

        Parameters
        ----------
        forward : Dataset
            The dataset to be thinned.
        thinning : Optional[int]
            The thinning factor.
        method : str
            The thinning method.
        """
        self.thinning = thinning
        self.method = method

        if thinning is not None:

            shape = forward.field_shape
            if len(shape) != 2:
                raise ValueError("Thinning only works latitude/longitude fields")

            # Make a copy, so we read the data only once from zarr
            forward_latitudes = forward.latitudes.copy()
            forward_longitudes = forward.longitudes.copy()

            latitudes = forward_latitudes.reshape(shape)
            longitudes = forward_longitudes.reshape(shape)
            latitudes = latitudes[::thinning, ::thinning].flatten()
            longitudes = longitudes[::thinning, ::thinning].flatten()

            # TODO: This is not very efficient

            mask = [lat in latitudes and lon in longitudes for lat, lon in zip(forward_latitudes, forward_longitudes)]
            mask = np.array(mask, dtype=bool)
        else:
            mask = None

        super().__init__(forward, mask)

    def mutate(self) -> Dataset:
        """Mutate the dataset.

        Returns
        -------
        Dataset
            The mutated dataset.
        """
        if self.thinning is None:
            return self.forward.mutate()
        return super().mutate()

    def tree(self) -> Node:
        """Get the tree representation of the dataset.

        Returns
        -------
        Node
            The tree representation of the dataset.
        """
        return Node(self, [self.forward.tree()], thinning=self.thinning, method=self.method)

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Get the metadata specific to the Thinning subclass.

        Returns
        -------
        Dict[str, Any]
            The metadata specific to the Thinning subclass.
        """
        return dict(thinning=self.thinning, method=self.method)


class Cropping(Masked):
    """A class to represent a cropped dataset."""

    def __init__(self, forward: Dataset, area: Union[Dataset, Tuple[float, float, float, float]]) -> None:
        """Initialize the Cropping class.

        Parameters
        ----------
        forward : Dataset
            The dataset to be cropped.
        area : Union[Dataset, Tuple[float, float, float, float]]
            The cropping area.
        """
        from ..data import open_dataset

        area = area if isinstance(area, (list, tuple)) else open_dataset(area)

        if isinstance(area, Dataset):
            north = np.amax(area.latitudes)
            south = np.amin(area.latitudes)
            east = np.amax(area.longitudes)
            west = np.amin(area.longitudes)
            area = (north, west, south, east)

        self.area = area
        mask = cropping_mask(forward.latitudes, forward.longitudes, *area)

        super().__init__(forward, mask)

    def tree(self) -> Node:
        """Get the tree representation of the dataset.

        Returns
        -------
        Node
            The tree representation of the dataset.
        """
        return Node(self, [self.forward.tree()], area=self.area)

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Get the metadata specific to the Cropping subclass.

        Returns
        -------
        Dict[str, Any]
            The metadata specific to the Cropping subclass.
        """
        return dict(area=self.area)


class TrimEdge(Masked):
    """A class that removes the boundary of a dataset."""

    def __init__(self, forward, edge):
        if isinstance(edge, int):
            self.edge = [edge] * 4
        elif isinstance(edge, (list, tuple)) and len(edge) == 4:
            self.edge = edge
        else:
            raise ValueError("'edge' must be an integer or a list of 4 integers")

        for e in self.edge:
            if not isinstance(e, int) or e < 0:
                raise ValueError("'edge' must be integer(s) 0 or greater")

        shape = forward.field_shape
        if len(shape) != 2:
            raise ValueError("TrimEdge only works on regular grids")

        if self.edge[0] + self.edge[1] >= shape[0]:
            raise ValueError("Too much triming of the first grid dimension, resulting in an empty dataset")
        if self.edge[2] + self.edge[3] >= shape[1]:
            raise ValueError("Too much triming of the second grid dimension, resulting in an empty dataset")

        mask = np.full(shape, True, dtype=bool)
        mask[0 : self.edge[0], :] = False
        mask[:, 0 : self.edge[2]] = False
        if self.edge[1] != 0:
            mask[-self.edge[1] :, :] = False
        if self.edge[3] != 0:
            mask[:, -self.edge[3] :] = False

        mask = mask.flatten()

        super().__init__(forward, mask)

    def mutate(self) -> Dataset:
        """Mutate the dataset.

        Returns
        -------
        Dataset
            The mutated dataset.
        """
        if self.edge is None:
            return self.forward.mutate()
        return super().mutate()

    def tree(self) -> Node:
        """Get the tree representation of the dataset.

        Returns
        -------
        Node
            The tree representation of the dataset.
        """
        return Node(self, [self.forward.tree()], edge=self.edge)

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Get the metadata specific to the TrimEdge subclass.

        Returns
        -------
        Dict[str, Any]
            The metadata specific to the TrimEdge subclass.
        """
        return dict(edge=self.edge)

    @property
    def field_shape(self) -> Shape:
        """Returns the field shape of the dataset."""
        x, y = self.forward.field_shape
        x -= self.edge[0] + self.edge[1]
        y -= self.edge[2] + self.edge[3]
        return x, y



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property
from typing import Any
from typing import Dict

from earthkit.data import FieldList

from ...dates.groups import GroupOfDates
from .action import Action
from .misc import _tidy
from .misc import assert_fieldlist
from .result import Result
from .template import notify_result
from .template import substitute
from .trace import trace
from .trace import trace_datasource
from .trace import trace_select

LOG = logging.getLogger(__name__)


class FunctionContext:
    """A FunctionContext is passed to all functions, it will be used to pass information
    to the functions from the other actions and filters and results.
    """

    def __init__(self, owner: Result) -> None:
        """Initializes a FunctionContext instance.

        Parameters
        ----------
        owner : object
            The owner object.
        """
        self.owner = owner
        self.use_grib_paramid: bool = owner.context.use_grib_paramid

    def trace(self, emoji: str, *args: Any) -> None:
        """Traces the given arguments with an emoji.

        Parameters
        ----------
        emoji : str
            The emoji to use.
        *args : Any
            The arguments to trace.
        """
        trace(emoji, *args)

    def info(self, *args: Any, **kwargs: Any) -> None:
        """Logs an info message.

        Parameters
        ----------
        *args : Any
            The arguments for the log message.
        **kwargs : Any
            The keyword arguments for the log message.
        """
        LOG.info(*args, **kwargs)

    @property
    def dates_provider(self) -> object:
        """Returns the dates provider."""
        return self.owner.group_of_dates.provider

    @property
    def partial_ok(self) -> bool:
        """Returns whether partial results are acceptable."""
        return self.owner.group_of_dates.partial_ok

    def get_result(self, *args, **kwargs) -> Any:
        return self.owner.context.get_result(*args, **kwargs)


class FunctionAction(Action):
    """Represents an action that executes a function.

    Attributes
    ----------
    name : str
        The name of the function.
    """

    def __init__(self, context: object, action_path: list, _name: str, source, **kwargs: Dict[str, Any]) -> None:
        """Initializes a FunctionAction instance.

        Parameters
        ----------
        context : object
            The context object.
        action_path : list
            The action path.
        _name : str
            The name of the function.
        **kwargs : Dict[str, Any]
            Additional keyword arguments.
        """
        super().__init__(context, action_path, **kwargs)
        self.name: str = _name
        self.source = source

    @trace_select
    def select(self, group_of_dates: GroupOfDates) -> "FunctionResult":
        """Selects the function result for the given group of dates.

        Parameters
        ----------
        group_of_dates : GroupOfDates
            The group of dates.

        Returns
        -------
        FunctionResult
            The function result instance.
        """
        return FunctionResult(self.context, self.action_path, group_of_dates, action=self)

    def __repr__(self) -> str:
        """Returns a string representation of the FunctionAction instance."""
        content: str = ""
        content += ",".join([self._short_str(a) for a in self.args])
        content += " ".join([self._short_str(f"{k}={v}") for k, v in self.kwargs.items()])
        content = self._short_str(content)
        return self._repr(_inline_=content, _indent_=" ")

    def _trace_select(self, group_of_dates: GroupOfDates) -> str:
        """Traces the selection of the function for the given group of dates.

        Parameters
        ----------
        group_of_dates : GroupOfDates
            The group of dates.

        Returns
        -------
        str
            The trace string.
        """
        return f"{self.name}({group_of_dates})"


class FunctionResult(Result):
    """Represents the result of executing a function.

    Attributes
    ----------
    action : Action
        The action instance.
    args : tuple
        The positional arguments for the function.
    kwargs : dict
        The keyword arguments for the function.
    """

    def __init__(self, context: object, action_path: list, group_of_dates: GroupOfDates, action: Action) -> None:
        """Initializes a FunctionResult instance.

        Parameters
        ----------
        context : object
            The context object.
        action_path : list
            The action path.
        group_of_dates : GroupOfDates
            The group of dates.
        action : Action
            The action instance.
        """
        super().__init__(context, action_path, group_of_dates)
        assert isinstance(action, Action), type(action)
        self.action: Action = action

        self.args, self.kwargs = substitute(context, (self.action.args, self.action.kwargs))

    def _trace_datasource(self, *args: Any, **kwargs: Any) -> str:
        """Traces the datasource for the given arguments.

        Parameters
        ----------
        *args : Any
            The arguments.
        **kwargs : Any
            The keyword arguments.

        Returns
        -------
        str
            The trace string.
        """
        return f"{self.action.name}({self.group_of_dates})"

    @cached_property
    @assert_fieldlist
    @notify_result
    @trace_datasource
    def datasource(self) -> FieldList:
        """Returns the datasource for the function result."""
        # args, kwargs = resolve(self.context, (self.args, self.kwargs))
        self.action.source.context = FunctionContext(self)

        return _tidy(
            self.action.source.execute(
                list(self.group_of_dates),  # Will provide a list of datetime objects
            )
        )

    def __repr__(self) -> str:
        """Returns a string representation of the FunctionResult instance."""
        try:
            return f"{self.action.name}({self.group_of_dates})"
        except Exception:
            return f"{self.__class__.__name__}(unitialised)"

    @property
    def function(self) -> None:
        """Raises NotImplementedError as this property is not implemented.

        Raises
        ------
        NotImplementedError
            Always raised.
        """
        raise NotImplementedError(f"Not implemented in {self.__class__.__name__}")



from anemoi.datasets import open_dataset

ds = open_dataset(
    dataset={
        "join": [
            {
                "dataset": "dataset-3h",
                "frequency": "3h",
            },
            {
                "dataset": "dataset-24h",
                "interpolate_frequency": "3h",
            },
        ],
        "adjust": "dates",
    },
    start="2004-01-01",
    end="2023-01-01",
)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import wraps
from typing import Any
from typing import Callable
from typing import Tuple
from typing import Union

from earthkit.data import FieldList
from earthkit.data.core.fieldlist import MultiFieldList

LOG = logging.getLogger(__name__)


def parse_function_name(name: str) -> Tuple[str, Union[int, None]]:
    """Parses a function name to extract the base name and an optional time delta.

    Parameters
    ----------
    name : str
        The function name to parse.

    Returns
    -------
    tuple of (str, int or None)
        The base name and an optional time delta.
    """
    if name.endswith("h") and name[:-1].isdigit():

        if "-" in name:
            name, delta = name.split("-")
            sign = -1

        elif "+" in name:
            name, delta = name.split("+")
            sign = 1

        else:
            return name, None

        assert delta[-1] == "h", (name, delta)
        delta = sign * int(delta[:-1])
        return name, delta

    return name, None


def assert_fieldlist(method: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to assert that the result of a method is an instance of FieldList.

    Parameters
    ----------
    method : Callable[..., Any]
        The method to decorate.

    Returns
    -------
    Callable[..., Any]
        The decorated method.
    """

    @wraps(method)
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:

        result = method(self, *args, **kwargs)
        assert isinstance(result, FieldList), type(result)
        return result

    return wrapper


def assert_is_fieldlist(obj: object) -> None:
    """Asserts that the given object is an instance of FieldList.

    Parameters
    ----------
    obj : object
        The object to check.
    """
    assert isinstance(obj, FieldList), type(obj)


def _flatten(ds: Union[MultiFieldList, FieldList]) -> list:
    """Flattens a MultiFieldList or FieldList into a list of FieldList objects.

    Parameters
    ----------
    ds : Union[MultiFieldList, FieldList]
        The dataset to flatten.

    Returns
    -------
    list
        A list of FieldList objects.
    """
    if isinstance(ds, MultiFieldList):
        return [_tidy(f) for s in ds._indexes for f in _flatten(s)]
    return [ds]


def _tidy(ds: Union[MultiFieldList, FieldList], indent: int = 0) -> Union[MultiFieldList, FieldList]:
    """Tidies up a MultiFieldList or FieldList by removing empty sources.

    Parameters
    ----------
    ds : Union[MultiFieldList, FieldList]
        The dataset to tidy.
    indent : int, optional
        The indentation level. Defaults to 0.

    Returns
    -------
    Union[MultiFieldList, FieldList]
        The tidied dataset.
    """
    if isinstance(ds, MultiFieldList):

        sources = [s for s in _flatten(ds) if len(s) > 0]
        if len(sources) == 1:
            return sources[0]
        return MultiFieldList(sources)
    return ds



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

import earthkit.data as ekd
import numpy as np
from anemoi.transform.fields import new_field_from_numpy
from anemoi.transform.fields import new_fieldlist_from_list
from earthkit.meteo import constants
from earthkit.meteo import thermo
from numpy.typing import NDArray

from .legacy import legacy_filter


# Alternative proposed by Baudouin Raoult
class AutoDict(dict):
    """A dictionary that automatically creates nested dictionaries for missing keys."""

    def __missing__(self, key: Any) -> Any:
        """Handle missing keys by creating nested dictionaries.

        Parameters
        ----------
        key : Any
            The missing key.

        Returns
        -------
        Any
            A new nested dictionary.
        """
        value = self[key] = type(self)()
        return value


def model_level_pressure(
    A: NDArray[Any], B: NDArray[Any], surface_pressure: Union[float, np.ndarray]
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Calculates:
     - pressure at the model full- and half-levels
     - delta: depth of log(pressure) at full levels
     - alpha: alpha term #TODO: more descriptive information.

    Parameters
    ----------
    A : ndarray
        A-coefficients defining the model levels
    B : ndarray
        B-coefficients defining the model levels
    surface_pressure : number or ndarray
        surface pressure (Pa)

    Returns
    -------
    ndarray
        pressure at model full-levels
    ndarray
        pressure at model half-levels
    ndarray
        delta at full-levels
    ndarray
        alpha at full levels
    """

    # constants
    PRESSURE_TOA = 0.1  # safety when highest pressure level = 0.0

    # make the calculation agnostic to the number of dimensions
    ndim = surface_pressure.ndim
    new_shape_half = (A.shape[0],) + (1,) * ndim
    A_reshaped = A.reshape(new_shape_half)
    B_reshaped = B.reshape(new_shape_half)

    # calculate pressure on model half-levels
    p_half_level = A_reshaped + B_reshaped * surface_pressure[np.newaxis, ...]

    # calculate delta
    new_shape_full = (A.shape[0] - 1,) + surface_pressure.shape
    delta = np.zeros(new_shape_full)
    delta[1:, ...] = np.log(p_half_level[2:, ...] / p_half_level[1:-1, ...])

    # pressure at highest half level<= 0.1
    if np.any(p_half_level[0, ...] <= PRESSURE_TOA):
        delta[0, ...] = np.log(p_half_level[1, ...] / PRESSURE_TOA)
    # pressure at highest half level > 0.1
    else:
        delta[0, ...] = np.log(p_half_level[1, ...] / p_half_level[0, ...])

    # calculate alpha
    alpha = np.zeros(new_shape_full)

    alpha[1:, ...] = 1.0 - p_half_level[1:-1, ...] / (p_half_level[2:, ...] - p_half_level[1:-1, ...]) * delta[1:, ...]

    # pressure at highest half level <= 0.1
    if np.any(p_half_level[0, ...] <= PRESSURE_TOA):
        alpha[0, ...] = 1.0  # ARPEGE choice, ECMWF IFS uses log(2)
    # pressure at highest half level > 0.1
    else:
        alpha[0, ...] = 1.0 - p_half_level[0, ...] / (p_half_level[1, ...] - p_half_level[0, ...]) * delta[0, ...]

    # calculate pressure on model full levels
    # TODO: is there a faster way to calculate the averages?
    # TODO: introduce option to calculate full levels in more complicated way
    p_full_level = np.apply_along_axis(lambda m: np.convolve(m, np.ones(2) / 2, mode="valid"), axis=0, arr=p_half_level)

    return p_full_level, p_half_level, delta, alpha


def calc_specific_gas_constant(q: Union[float, np.ndarray]) -> Union[float, NDArray[Any]]:
    """Calculates the specific gas constant of moist air
    (specific content of cloud particles and hydrometeors are neglected).

    Parameters
    ----------
    q : number or ndarray
        specific humidity

    Returns
    -------
    number or ndarray
        specific gas constant of moist air
    """

    R = constants.Rd + (constants.Rv - constants.Rd) * q
    return R


def relative_geopotential_thickness(alpha: NDArray[Any], q: NDArray[Any], T: NDArray[Any]) -> NDArray[Any]:
    """Calculates the geopotential thickness w.r.t the surface on model full-levels.

    Parameters
    ----------
    alpha : ndarray
        alpha term of pressure calculations
    q : ndarray
        specific humidity (in kg/kg) on model full-levels
    T : ndarray
        temperature (in Kelvin) on model full-levels

    Returns
    -------
    ndarray
        geopotential thickness of model full-levels w.r.t. the surface
    """

    R = calc_specific_gas_constant(q)
    dphi = np.cumsum(np.flip(alpha * R * T, axis=0), axis=0)
    dphi = np.flip(dphi, axis=0)

    return dphi


def pressure_at_height_level(
    height: float, q: NDArray[Any], T: NDArray[Any], sp: NDArray[Any], A: NDArray[Any], B: NDArray[Any]
) -> Union[float, NDArray[Any]]:
    """Calculates the pressure at a height level given in meters above surface.
    This is done by finding the model level above and below the specified height
    and interpolating the pressure.

    Parameters
    ----------
    height : number
        height (in meters) above the surface for which the pressure is wanted
    q : ndarray
        specific humidity (kg/kg) at model full-levels
    T : ndarray
        temperature (K) at model full-levels
    sp : ndarray
        surface pressure (Pa)
    A : ndarray
        A-coefficients defining the model levels
    B : ndarray
        B-coefficients defining the model levels

    Returns
    -------
    number or ndarray
        pressure (Pa) at the given height level
    """

    # geopotential thickness of the height level
    tdphi = height * constants.g

    # pressure(-related) variables
    p_full, p_half, _, alpha = model_level_pressure(A, B, sp)

    # relative geopot. thickness of full levels
    dphi = relative_geopotential_thickness(alpha, q, T)

    # find the model full level right above the height level
    i_phi = (tdphi > dphi).sum(0)

    # initialize the output array
    p_height = np.zeros_like(i_phi, dtype=np.float64)

    # define mask: requested height is below the lowest model full-level
    mask = i_phi == 0

    # CASE 1: requested height is below the lowest model full-level
    # --> interpolation between surface pressure and lowest model full-level
    p_height[mask] = (p_half[-1, ...] + tdphi / dphi[-1, ...] * (p_full[-1, ...] - p_half[-1, ...]))[mask]

    # CASE 2: requested height is above the lowest model full-level
    # --> interpolation between between model full-level above and below

    # define some indices for masking and readability
    i_lev = alpha.shape[0] - i_phi - 1  # convert phi index to model level index
    indices = np.indices(i_lev.shape)
    masked_indices = tuple(dim[~mask] for dim in indices)
    above = (i_lev[~mask],) + masked_indices
    below = (i_lev[~mask] + 1,) + masked_indices

    dphi_above = dphi[above]
    dphi_below = dphi[below]

    factor = (tdphi - dphi_above) / (dphi_below - dphi_above)
    p_height[~mask] = p_full[above] + factor * (p_full[below] - p_full[above])

    return p_height


@legacy_filter(__file__)
def execute(
    context: Any,
    input: List[Any],
    height: float,
    t: str,
    q: str,
    sp: str,
    new_name: str = "2r",
    **kwargs: Dict[str, Any],
) -> ekd.FieldList:
    """Convert the single (height) level specific humidity to relative humidity.

    Parameters
    ----------
    context : Any
        The context for the execution.
    input : list of Any
        The input data.
    height : float
        The height level in meters.
    t : str
        The temperature parameter name.
    q : str
        The specific humidity parameter name.
    sp : str
        The surface pressure parameter name.
    new_name : str, optional
        The new name for the relative humidity parameter, by default "2r".
    **kwargs : dict
        Additional keyword arguments.
        t_ml : str, optional
            The temperature parameter name for model levels, by default "t".
        q_ml : str, optional
            The specific humidity parameter name for model levels, by default "q".
        A : list of float
            A-coefficients defining the model levels.
        B : list of float
            B-coefficients defining the model levels.
        keep_q : bool, optional
            Whether to keep the specific humidity field in the result, by default False.

    Returns
    -------
    ekd.FieldList
        The resulting field array with relative humidity.
    """
    result = []

    MANDATORY_KEYS = ["A", "B"]
    OPTIONAL_KEYS = ["t_ml", "q_ml"]
    MISSING_KEYS = []
    DEFAULTS = dict(t_ml="t", q_ml="q")

    for key in OPTIONAL_KEYS:
        if key not in kwargs:
            print(f"key {key} not found in yaml-file, using default key: {DEFAULTS[key]}")
            kwargs[key] = DEFAULTS[key]

    for key in MANDATORY_KEYS:
        if key not in kwargs:
            MISSING_KEYS.append(key)

    if MISSING_KEYS:
        raise KeyError(f"Following keys are missing: {', '.join(MISSING_KEYS)}")

    single_level_params = (t, q, sp)
    model_level_params = (kwargs["t_ml"], kwargs["q_ml"])

    needed_fields = AutoDict()

    # Gather all necessary fields
    for f in input:
        key = f.metadata(namespace="mars")
        param = key.pop("param")
        # check single level parameters
        if param in single_level_params:
            levtype = key.pop("levtype")
            key = tuple(sorted(key.items()))

            if param in needed_fields[key][levtype]:
                raise ValueError(f"Duplicate single level field {param} for {key}")

            needed_fields[key][levtype][param] = f
            if param == q:
                if kwargs.get("keep_q", False):
                    result.append(f)
            else:
                result.append(f)

        # check model level parameters
        elif param in model_level_params:
            levtype = key.pop("levtype")
            levelist = key.pop("levelist")
            key = tuple(sorted(key.items()))

            if param in needed_fields[key][levtype][levelist]:
                raise ValueError(f"Duplicate model level field {param} for {key} at level {levelist}")

            needed_fields[key][levtype][levelist][param] = f

        # all other parameters
        else:
            result.append(f)

    for _, values in needed_fields.items():
        # some checks
        if len(values["sfc"]) != 3:
            raise ValueError("Missing surface fields")

        q_sl = values["sfc"][q].to_numpy(flatten=True)
        t_sl = values["sfc"][t].to_numpy(flatten=True)
        sp_sl = values["sfc"][sp].to_numpy(flatten=True)

        nlevels = len(kwargs["A"]) - 1
        if len(values["ml"]) != nlevels:
            raise ValueError("Missing model levels")

        for key in values["ml"].keys():
            if len(values["ml"][key]) != 2:
                raise ValueError(f"Missing field on level {key}")

        # create 3D arrays for upper air fields
        levels = list(values["ml"].keys())
        levels.sort()
        t_ml = []
        q_ml = []
        for level in levels:
            t_ml.append(values["ml"][level][kwargs["t_ml"]].to_numpy(flatten=True))
            q_ml.append(values["ml"][level][kwargs["q_ml"]].to_numpy(flatten=True))

        t_ml = np.stack(t_ml)
        q_ml = np.stack(q_ml)

        # actual conversion from qv --> rh
        # FIXME:
        # For now We need to go from qv --> td --> rh to take into account
        # the mixed / ice phase when T ~ 0C / T < 0C
        # See https://github.com/ecmwf/earthkit-meteo/issues/15
        p_sl = pressure_at_height_level(height, q_ml, t_ml, sp_sl, np.array(kwargs["A"]), np.array(kwargs["B"]))
        td_sl = thermo.dewpoint_from_specific_humidity(q=q_sl, p=p_sl)
        rh_sl = thermo.relative_humidity_from_dewpoint(t=t_sl, td=td_sl)

        result.append(new_field_from_numpy(values["sfc"][q], rh_sl, param=new_name))

    return new_fieldlist_from_list(result)


def test() -> None:
    """Test the conversion from specific humidity to relative humidity.

    This function fetches data from a source, performs the conversion, and prints
    the mean, median, and maximum differences in dewpoint temperature.

    Returns
    -------
    None
    """
    from earthkit.data import from_source
    from earthkit.data.readers.grib.index import GribFieldList

    # IFS forecasts have both specific humidity and dewpoint
    sl = from_source(
        "mars",
        {
            "date": "2022-01-01",
            "class": "od",
            "expver": "1",
            "stream": "oper",
            "levtype": "sfc",
            "param": "96.174/134.128/167.128/168.128",
            "time": "00:00:00",
            "type": "fc",
            "step": "2",
            "grid": "O640",
        },
    )

    ml = from_source(
        "mars",
        {
            "date": "2022-01-01",
            "class": "od",
            "expver": "1",
            "stream": "oper",
            "levtype": "ml",
            "levelist": "130/131/132/133/134/135/136/137",
            "param": "130/133",
            "time": "00:00:00",
            "type": "fc",
            "step": "2",
            "grid": "O640",
        },
    )
    source = GribFieldList.merge([sl, ml])

    # IFS A and B coeffients for level 137 - 129
    kwargs = {
        "A": [424.414063, 302.476563, 202.484375, 122.101563, 62.781250, 22.835938, 3.757813, 0.0, 0.0],
        "B": [0.969513, 0.975078, 0.980072, 0.984542, 0.988500, 0.991984, 0.995003, 0.997630, 1.000000],
    }
    source = execute(None, source, 2, "2t", "2sh", "sp", "2r", **kwargs)

    temperature = source[2].to_numpy(flatten=True)
    dewpoint = source[3].to_numpy(flatten=True)
    relhum = source[4].to_numpy()
    newdew = thermo.dewpoint_from_relative_humidity(temperature, relhum)

    print(f"Mean difference in dewpoint temperature: {np.abs(newdew - dewpoint).mean():02f} degC")
    print(f"Median difference in dewpoint temperature: {np.median(np.abs(newdew - dewpoint)):02f} degC")
    print(f"Maximum difference in dewpoint temperature: {np.abs(newdew - dewpoint).max():02f} degC")

    # source.save("source.grib")


if __name__ == "__main__":
    test()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
import os
import warnings
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Union
from urllib.parse import urlparse

import numpy as np
import zarr
from anemoi.utils.dates import frequency_to_timedelta
from numpy.typing import NDArray

from . import MissingDateError
from .dataset import Dataset
from .dataset import FullIndex
from .dataset import Shape
from .dataset import TupleIndex
from .debug import DEBUG_ZARR_LOADING
from .debug import Node
from .debug import Source
from .debug import debug_indexing
from .indexing import expand_list_indexing
from .misc import load_config

LOG = logging.getLogger(__name__)


class ReadOnlyStore(zarr.storage.BaseStore):
    """A base class for read-only stores."""

    def __delitem__(self, key: str) -> None:
        """Prevent deletion of items."""
        raise NotImplementedError()

    def __setitem__(self, key: str, value: bytes) -> None:
        """Prevent setting of items."""
        raise NotImplementedError()

    def __len__(self) -> int:
        """Return the number of items in the store."""
        raise NotImplementedError()

    def __iter__(self) -> iter:
        """Return an iterator over the store."""
        raise NotImplementedError()


class HTTPStore(ReadOnlyStore):
    """A read-only store for HTTP(S) resources."""

    def __init__(self, url: str) -> None:
        """Initialize the HTTPStore with a URL."""
        self.url = url

    def __getitem__(self, key: str) -> bytes:
        """Retrieve an item from the store."""
        import requests

        r = requests.get(self.url + "/" + key)

        if r.status_code == 404:
            raise KeyError(key)

        r.raise_for_status()
        return r.content


class S3Store(ReadOnlyStore):
    """A read-only store for S3 resources."""

    """We write our own S3Store because the one used by zarr (s3fs)
    does not play well with fork(). We also get to control the s3 client
    options using the anemoi configs.
    """

    def __init__(self, url: str, region: Optional[str] = None) -> None:
        """Initialize the S3Store with a URL and optional region."""
        from anemoi.utils.remote.s3 import s3_client

        _, _, self.bucket, self.key = url.split("/", 3)
        self.s3 = s3_client(self.bucket, region=region)

    def __getitem__(self, key: str) -> bytes:
        """Retrieve an item from the store."""
        try:
            response = self.s3.get_object(Bucket=self.bucket, Key=self.key + "/" + key)
        except self.s3.exceptions.NoSuchKey:
            raise KeyError(key)

        return response["Body"].read()


class PlanetaryComputerStore(ReadOnlyStore):
    """We write our own Store to access catalogs on Planetary Computer,
    as it requires some extra arguments to use xr.open_zarr.
    """

    def __init__(self, data_catalog_id: str) -> None:
        """Initialize the PlanetaryComputerStore with a data catalog ID.

        Parameters
        ----------
        data_catalog_id : str
            The data catalog ID.
        """
        self.data_catalog_id = data_catalog_id

        import planetary_computer
        import pystac_client

        catalog = pystac_client.Client.open(
            "https://planetarycomputer.microsoft.com/api/stac/v1/",
            modifier=planetary_computer.sign_inplace,
        )
        collection = catalog.get_collection(self.data_catalog_id)

        asset = collection.assets["zarr-abfs"]

        if "xarray:storage_options" in asset.extra_fields:
            store = {
                "store": asset.href,
                "storage_options": asset.extra_fields["xarray:storage_options"],
                **asset.extra_fields["xarray:open_kwargs"],
            }
        else:
            store = {
                "filename_or_obj": asset.href,
                **asset.extra_fields["xarray:open_kwargs"],
            }

        self.store = store

    def __getitem__(self, key: str) -> bytes:
        """Retrieve an item from the store."""
        raise NotImplementedError()


class DebugStore(ReadOnlyStore):
    """A store to debug the zarr loading."""

    def __init__(self, store: ReadOnlyStore) -> None:
        """Initialize the DebugStore with another store."""
        assert not isinstance(store, DebugStore)
        self.store = store

    def __getitem__(self, key: str) -> bytes:
        """Retrieve an item from the store and print debug information."""
        # print()
        print("GET", key, self)
        # traceback.print_stack(file=sys.stdout)
        return self.store[key]

    def __len__(self) -> int:
        """Return the number of items in the store."""
        return len(self.store)

    def __iter__(self) -> iter:
        """Return an iterator over the store."""
        warnings.warn("DebugStore: iterating over the store")
        return iter(self.store)

    def __contains__(self, key: str) -> bool:
        """Check if the store contains a key."""
        return key in self.store


def name_to_zarr_store(path_or_url: str) -> ReadOnlyStore:
    """Convert a path or URL to a zarr store."""
    store = path_or_url

    if store.startswith("s3://"):
        store = S3Store(store)

    elif store.startswith("http://") or store.startswith("https://"):
        parsed = urlparse(store)
        bits = parsed.netloc.split(".")
        if len(bits) == 5 and (bits[1], bits[3], bits[4]) == ("s3", "amazonaws", "com"):
            s3_url = f"s3://{bits[0]}{parsed.path}"
            store = S3Store(s3_url, region=bits[2])
        elif store.startswith("https://planetarycomputer.microsoft.com/"):
            data_catalog_id = store.rsplit("/", 1)[-1]
            store = PlanetaryComputerStore(data_catalog_id).store
        else:
            store = HTTPStore(store)

    return store


def open_zarr(path: str, dont_fail: bool = False, cache: int = None) -> zarr.hierarchy.Group:
    """Open a zarr store from a path."""
    try:
        store = name_to_zarr_store(path)

        if DEBUG_ZARR_LOADING:
            if isinstance(store, str):
                import os

                if not os.path.isdir(store):
                    raise NotImplementedError(
                        "DEBUG_ZARR_LOADING is only implemented for DirectoryStore. "
                        "Please disable it for other backends."
                    )
                store = zarr.storage.DirectoryStore(store)
            store = DebugStore(store)

        if cache is not None:
            store = zarr.LRUStoreCache(store, max_size=cache)

        return zarr.convenience.open(store, "r")
    except zarr.errors.PathNotFoundError:
        if not dont_fail:
            raise zarr.errors.PathNotFoundError(path)


class Zarr(Dataset):
    """A zarr dataset."""

    def __init__(self, path: Union[str, zarr.hierarchy.Group]) -> None:
        """Initialize the Zarr dataset with a path or zarr group."""
        if isinstance(path, zarr.hierarchy.Group):
            self.was_zarr = True
            self.path = str(id(path))
            self.z = path
        else:
            self.was_zarr = False
            self.path = str(path)
            self.z = open_zarr(self.path)

        # This seems to speed up the reading of the data a lot
        self.data = self.z.data
        self._missing = set()

    @property
    def missing(self) -> Set[int]:
        """Return the missing dates of the dataset."""
        return self._missing

    @classmethod
    def from_name(cls, name: str) -> "Zarr":
        """Create a Zarr dataset from a name."""
        if name.endswith(".zip") or name.endswith(".zarr"):
            return Zarr(name)
        return Zarr(zarr_lookup(name))

    def __len__(self) -> int:
        """Return the length of the dataset."""
        return self.data.shape[0]

    @debug_indexing
    @expand_list_indexing
    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Retrieve an item from the dataset."""
        return self.data[n]

    def _unwind(self, index: Union[int, slice, list, tuple], rest: list, shape: tuple, axis: int, axes: list) -> iter:
        """Unwind the index for multi-dimensional indexing."""
        if not isinstance(index, (int, slice, list, tuple)):
            try:
                # NumPy arrays, TensorFlow tensors, etc.
                index = tuple(index.tolist())
                assert not isinstance(index, bool), "Mask not supported"
            except AttributeError:
                pass

        if isinstance(index, (list, tuple)):
            axes.append(axis)  # Dimension of the concatenation
            for i in index:
                yield from self._unwind((slice(i, i + 1),), rest, shape, axis, axes)
            return

        if len(rest) == 0:
            yield (index,)
            return

        for n in self._unwind(rest[0], rest[1:], shape, axis + 1, axes):
            yield (index,) + n

    @cached_property
    def chunks(self) -> TupleIndex:
        """Return the chunks of the dataset."""
        return self.z.data.chunks

    @cached_property
    def shape(self) -> Shape:
        """Return the shape of the dataset."""
        return self.data.shape

    @cached_property
    def dtype(self) -> np.dtype:
        """Return the data type of the dataset."""
        return self.z.data.dtype

    @cached_property
    def dates(self) -> NDArray[np.datetime64]:
        """Return the dates of the dataset."""
        return self.z.dates[:]  # Convert to numpy

    @property
    def latitudes(self) -> NDArray[Any]:
        """Return the latitudes of the dataset."""
        try:
            return self.z.latitudes[:]
        except AttributeError:
            LOG.warning("No 'latitudes' in %r, trying 'latitude'", self)
            return self.z.latitude[:]

    @property
    def longitudes(self) -> NDArray[Any]:
        """Return the longitudes of the dataset."""
        try:
            return self.z.longitudes[:]
        except AttributeError:
            LOG.warning("No 'longitudes' in %r, trying 'longitude'", self)
            return self.z.longitude[:]

    @property
    def statistics(self) -> Dict[str, NDArray[Any]]:
        """Return the statistics of the dataset."""
        return dict(
            mean=self.z.mean[:],
            stdev=self.z.stdev[:],
            maximum=self.z.maximum[:],
            minimum=self.z.minimum[:],
        )

    def statistics_tendencies(self, delta: Optional[datetime.timedelta] = None) -> Dict[str, NDArray[Any]]:
        """Return the statistical tendencies of the dataset."""
        if delta is None:
            delta = self.frequency
        if isinstance(delta, int):
            delta = f"{delta}h"
        from anemoi.utils.dates import frequency_to_string
        from anemoi.utils.dates import frequency_to_timedelta

        delta = frequency_to_timedelta(delta)
        delta = frequency_to_string(delta)

        def func(k: str) -> str:
            return f"statistics_tendencies_{delta}_{k}"

        return dict(
            mean=self.z[func("mean")][:],
            stdev=self.z[func("stdev")][:],
            maximum=self.z[func("maximum")][:],
            minimum=self.z[func("minimum")][:],
        )

    @property
    def resolution(self) -> str:
        """Return the resolution of the dataset."""
        return self.z.attrs["resolution"]

    @property
    def field_shape(self) -> tuple:
        """Return the field shape of the dataset."""
        try:
            return tuple(self.z.attrs["field_shape"])
        except KeyError:
            LOG.warning("No 'field_shape' in %r, assuming 1D fields", self)
            return (self.shape[-1],)

    @property
    def frequency(self) -> datetime.timedelta:
        """Return the frequency of the dataset."""
        try:
            return frequency_to_timedelta(self.z.attrs["frequency"])
        except KeyError:
            LOG.warning("No 'frequency' in %r, computing from 'dates'", self)
        dates = self.dates
        return dates[1].astype(object) - dates[0].astype(object)

    @property
    def name_to_index(self) -> Dict[str, int]:
        """Return the name to index mapping of the dataset."""
        if "variables" in self.z.attrs:
            return {n: i for i, n in enumerate(self.z.attrs["variables"])}
        return self.z.attrs["name_to_index"]

    @property
    def variables(self) -> List[str]:
        """Return the variables of the dataset."""
        return [
            k
            for k, v in sorted(
                self.name_to_index.items(),
                key=lambda x: x[1],
            )
        ]

    @cached_property
    def constant_fields(self) -> List[str]:
        """Return the constant fields of the dataset."""
        result = self.z.attrs.get("constant_fields")
        if result is None:
            LOG.warning("No 'constant_fields' attribute in %r, computing them", self)
        return self.computed_constant_fields()

    @property
    def variables_metadata(self) -> Dict[str, Any]:
        """Return the metadata of the variables."""
        return self.z.attrs.get("variables_metadata", {})

    def __repr__(self) -> str:
        """Return the string representation of the dataset."""
        return self.path

    def end_of_statistics_date(self) -> np.datetime64:
        """Return the end date of the statistics."""
        return self.dates[-1]

    def metadata_specific(self, **kwargs: Any) -> Dict[str, Any]:
        """Return the specific metadata of the dataset."""
        return super().metadata_specific(
            attrs=dict(self.z.attrs),
            chunks=self.chunks,
            dtype=str(self.dtype),
            path=self.path,
        )

    def source(self, index: int) -> Source:
        """Return the source of the dataset."""
        return Source(self, index, info=self.path)

    def mutate(self) -> Dataset:
        """Mutate the dataset if it has missing dates."""
        if len(self.z.attrs.get("missing_dates", [])):
            LOG.warning(f"Dataset {self} has missing dates")
            return ZarrWithMissingDates(self.z if self.was_zarr else self.path)
        return self

    def tree(self) -> Node:
        """Return the tree representation of the dataset."""
        return Node(self, [], path=self.path)

    def get_dataset_names(self, names: Set[str]) -> None:
        """Get the names of the datasets."""
        name, _ = os.path.splitext(os.path.basename(self.path))
        names.add(name)

    def collect_supporting_arrays(self, collected: set, *path: str) -> None:
        """Collect supporting arrays."""
        pass

    def collect_input_sources(self, collected: set) -> None:
        """Collect input sources."""
        pass


class ZarrWithMissingDates(Zarr):
    """A zarr dataset with missing dates."""

    def __init__(self, path: Union[str, zarr.hierarchy.Group]) -> None:
        """Initialize the ZarrWithMissingDates dataset with a path or zarr group."""
        super().__init__(path)

        missing_dates = self.z.attrs.get("missing_dates", [])
        missing_dates = set([np.datetime64(x, "s") for x in missing_dates])
        self.missing_to_dates = {i: d for i, d in enumerate(self.dates) if d in missing_dates}
        self._missing = set(self.missing_to_dates)

    @property
    def missing(self) -> Set[int]:
        """Return the missing dates of the dataset."""
        return self._missing

    def mutate(self) -> Dataset:
        """Mutate the dataset."""
        return self

    @debug_indexing
    @expand_list_indexing
    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Retrieve an item from the dataset."""
        if isinstance(n, int):
            if n in self.missing:
                self._report_missing(n)
            return self.data[n]

        if isinstance(n, slice):
            common = set(range(*n.indices(len(self)))) & self.missing
            if common:
                self._report_missing(list(common)[0])
            return self.data[n]

        if isinstance(n, tuple):
            first = n[0]
            if isinstance(first, int):
                if first in self.missing:
                    self._report_missing(first)
                return self.data[n]

            if isinstance(first, slice):
                common = set(range(*first.indices(len(self)))) & self.missing
                if common:
                    self._report_missing(list(common)[0])
                return self.data[n]

            if isinstance(first, (list, tuple)):
                common = set(first) & self.missing
                if common:
                    self._report_missing(list(common)[0])
                return self.data[n]

            raise TypeError(f"Unsupported index {n} {type(n)}, {first} {type(first)}")

        raise TypeError(f"Unsupported index {n} {type(n)}")

    def _report_missing(self, n: int) -> None:
        """Report a missing date."""
        raise MissingDateError(f"Date {self.missing_to_dates[n]} is missing (index={n})")

    def tree(self) -> Node:
        """Return the tree representation of the dataset."""
        return Node(self, [], path=self.path, missing=sorted(self.missing))

    @property
    def label(self) -> str:
        """Return the label of the dataset."""
        return "zarr*"


QUIET = set()


def zarr_lookup(name: str, fail: bool = True) -> Optional[str]:
    """Look up a zarr dataset by name."""
    if name.endswith(".zarr") or name.endswith(".zip"):
        return name

    config = load_config()["datasets"]

    if name in config["named"]:
        if name not in QUIET:
            LOG.info("Opening `%s` as `%s`", name, config["named"][name])
            QUIET.add(name)
        return str(config["named"][name])

    tried = []
    for location in config["path"]:
        if not location.endswith("/"):
            location += "/"
        full = location + name + ".zarr"
        tried.append(full)
        try:
            z = open_zarr(full, dont_fail=True)
            if z is not None:
                # Cache for next time
                config["named"][name] = full
                if name not in QUIET:
                    LOG.info("Opening `%s` as `%s`", name, full)
                    QUIET.add(name)
                return full
        except zarr.errors.PathNotFoundError:
            pass

    if fail:
        raise ValueError(f"Cannot find a dataset that matched '{name}'. Tried: {tried}")

    return None
